/*---------------------------------------------------------------------------
 * 
 * Synopsis
 * --------
 * Conditionally-compiled wrappers to hide differences between BSD and 
 * Winsock sockets:
 * o initialization: none vs WSAStartup()
 * o termination:    none vs WSACleanup()
 * o closing:        close() vs closesocket()
 * o blocking:       fcntl() vs ioctlsocket()
 * o error handling  errno vs WSAGetLastError
 * o error codes:    EWOULDBLOCK vs WSAEWOULDBLOCK
 *
 * Compiling
 * ---------
 * For compiling under unix,  leave  _WIN32 undefined.
 * For compiling with Microsoft VC, define _WIN32.
 *
 * Notes
 * -----
 * I add support for various differences as I encounter them.
 *
 * Author
 * ------
 * Raymond McKendall
 * November 1998 -- January 1999
 *
 *--------------------------------------------------------------------------*/

#include "bsd-winsock.h"

#ifdef _WIN32

/*-----------------------------------------------------------------------------
 *
 * Winsock socket API
 *
 *---------------------------------------------------------------------------*/

int comm_initialize_sockets(void)
{
  WSADATA info;
  int rc;

  rc= WSAStartup(MAKEWORD(1,1), &info); /* Winsock 1.1 */
  if (rc == 0) {
    return(0);
  }
  else {
    fprintf(stderr, "comm_initialize_socket: WSAStartup() failed: ");
    if (rc == WSASYSNOTREADY) {
      fprintf(stderr, "Network is not ready.");
    }
    else  if (rc == WSAVERNOTSUPPORTED) {
      fprintf(stderr, "Implementation does not provide requested WinSock version.");
    }
    else  if (rc == WSAEINVAL) {
      fprintf(stderr, "DLL does not support requested WinSock version.");
    }
    else {
      fprintf(stderr, "unknown error.");
    }
    return(-1);
  }
}


int set_blocking(SOCKET sfd, int block)  /* Not tested. */
{
  u_long noblock = (block ? 0L : 1L);
  int rc;

  rc= ioctlsocket(sfd, FIONBIO, &noblock);

  if(rc) {errno = WSAGetLastError();}

  return(rc);
}
  

void comm_error(char* msg) {
  fprintf(stderr, "%s: Winsock error %d\n", msg, WSAGetLastError());
}

int comm_lasterror(void){return (WSAGetLastError());}

void comm_terminate_sockets(void)     {WSACleanup();}
void comm_close_socket(SOCKET sfd) {closesocket(sfd);}

#else

/*-----------------------------------------------------------------------------
 *
 * BSD socket API
 *
 *---------------------------------------------------------------------------*/

#include <unistd.h>  /* For close() under gcc with strict warnings. */
#include <fcntl.h>

int comm_initialize_sockets(void) {return(0);}
int comm_lasterror(void){return (errno);}
void comm_error(char * msg) {perror(msg);}
void comm_terminate_sockets(void) {}
void comm_close_socket(SOCKET sfd) {close(sfd);}

int set_blocking(SOCKET sfd, int block)   /* Not tested. */
{
  int rc;

  rc= (fcntl(sfd, F_SETFL, O_NONBLOCK) == -1);

  return(rc);
}
 


#endif

/*---------------------------------------------------------------------------*/
/*---- Test Module ----------------------------------------------------------
 * 
 * Synopsis
 * --------
 * Connects to echo server on specified host, then disconnects.
 *
 ---------------------------------------------------------------------------*/

#ifdef TEST
#undef TEST

#define CheckRC(msg) {if(rc) {fprintf(stderr, "%s\n", msg); exit(1);}}

#define ECHO_PORT 7

int main(int argc, char *argv[])
{
  HostEntry*      hostent;   /* Host entry for specified server. */
  SocketAddressIN address;   /* Socket address info for server.  */
  SOCKET sfd;                /* Socket to server.                */
  char* host;                /* Host name of server.             */
  int rc;

  /*--- Get hostname from command line. ---*/
  if (argc == 1+1) {
    host = *++argv;
  }
  else {
    fprintf(stderr,"usage: client <host name>\n");
    exit(1);
  }

  /*--- Initialize socket support ---*/
  rc= comm_initialize_sockets();   CheckRC("Cannot initialize sockets\n");

  /*--- Fill in socket address. ---*/
  memset(&address, 0, sizeof(SocketAddressIN));
  address.sin_family = AF_INET;
  address.sin_port   = htons((u_short) ECHO_PORT);
  hostent = gethostbyname(host); 
  if (hostent) { 
    address.sin_addr = *((NetAddress *) hostent->h_addr);
  }
  else {
    fprintf(stderr, "Cannot resolve host address\n");
    exit(1);
  }
  
  /*--- Create socket. ---*/
  sfd = socket(AF_INET, SOCK_STREAM, PROTOCOL);
  if (sfd == SOCKET_ERROR) {
    comm_error("Cannot create socket");
    exit(1);
  }

  /*--- Connect to server. ---*/
  rc= connect(sfd, (SocketAddress *)&address, sizeof(SocketAddress));
  if (rc == SOCKET_ERROR) {
    comm_error("Cannot connect to server");
    exit(1);
  }

  /*--- Toggle blocking. ---*/
  rc= set_blocking(sfd, 0);   CheckRC("Failed to set non-blocking mode.\n");
  rc= set_blocking(sfd, 1);   CheckRC("Failed to set blocking mode.\n");

  printf("OK\n");

  /*--- Clean up. ---*/
  comm_close_socket(sfd);
  comm_terminate_sockets();

  return(0);
}

#endif

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
