/*---------------------------------------------------------------------------
 * 
 * Synopsis
 * --------
 * Common utilities for clients.
 *
 * Author
 * ------
 * Raymond McKendall
 * November 1998
 *
 *--------------------------------------------------------------------------*/

#include <ctype.h> /* For isdigit() */
#include "comm.h"

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
int comm_engage_server(
   SOCKET* sfd,        /* Socket to create for server. */
   const char host[],  /* Either name or IP address.   */
   u_short port        /* Target server's port.        */
)
/*
 * Description
 * -----------
 * Creates socket connected to server specified by host and port.
 *
 * Note
 * ----
 * If the first character of host[] is a digit, then host[] is assumed
 * to specify an IP address. Otherwise, host[] is assumed to 
 * specify a name.
 *
 *---------------------------------------------------------------------------*/
{
  SocketAddressIN address;   /* Socket address info for server.  */
  u_long          ipaddress; /* Used if host[] is an IP address. */
  HostEntry*      hostent;   /* Used if host[] is a name.        */
  int rc;
 
  /*--- Initialize socket support ---*/
  rc= comm_initialize_sockets();
  if (rc) {
    fprintf(stderr, "comm_engage_server: Cannot initialize socket support\n");
    return(1);
  }

  /*--- Fill in socket address. ---*/
  memset(&address, 0, sizeof(SocketAddressIN));
  address.sin_family = AF_INET;
  address.sin_port   = htons(port);

  /*--- Get server's address. ---*/
  if (isdigit((int) host[0])) { /* If host[] is an IP address. */
    ipaddress = inet_addr(host);
    address.sin_addr = *((NetAddress *) &ipaddress);
  }
  else { /* Else host[] is a name. */
    hostent = gethostbyname(host); 
    if (hostent) { 
      address.sin_addr = *((NetAddress *) hostent->h_addr);
    }
    else {
      fprintf(stderr, "comm_engage_server: Cannot resolve host address\n");
      return(1);
    }
  }

  /*--- Create socket. ---*/
  *sfd = socket(AF_INET, SOCK_STREAM, PROTOCOL);
  if (*sfd == SOCKET_ERROR) {
    comm_error("comm_engage_server: Cannot create socket");
    return(1);
  }

  /*--- Connect to server. ---*/
  rc= connect(*sfd, (SocketAddress *)&address, sizeof(SocketAddress));
  if (rc == SOCKET_ERROR) {
    comm_error("comm_engage_server: Cannot connect to server");
    return(1);
  }

  return(0);
}

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
void comm_disengage_server(
   SOCKET sfd    /* Socket to server. */
)
/*
 * Description
 * -----------
 * Close server on socket sfd and terminates socket support.
 *
 *---------------------------------------------------------------------------*/
{
  comm_close_socket(sfd);
  comm_terminate_sockets();
}

/*---------------------------------------------------------------------------*/
/*---- Test Module ----------------------------------------------------------
 *
 * Synopsis
 * --------
 * Opens a connection to standard echo server on port 7 of specified host
 * and pings it with a single character.
 *
 *---------------------------------------------------------------------------*/

#ifdef TEST
#undef TEST

#define CheckRC(msg) {if(rc) {fprintf(stderr, "%s\n", msg); exit(1);}}

#define ECHO_PORT 7

int main(int argc, char *argv[])
{
  SOCKET sfd;             /* Socket to server.               */
  char* host;             /* Host name or address of server. */
  char ping_send = 'P';   /* Ping message sent.              */
  char ping_recv;         /* Ping message received.          */ 
  int rc;

  /*--- Get server's host from command line. ---*/
  if (argc == 1+1) {
    host = *++argv;
  }
  else {
    fprintf(stderr,"main: usage: %s <server host>\n", argv[0]);
    exit(1);
  }

  /*--- Engage server ---*/
  rc= comm_engage_server(&sfd, host, ECHO_PORT);
  CheckRC("Could not open server\n");
  printf("Connected to \"%s\" on port %d.\n", host, ECHO_PORT);

  /*--- Ping ---*/
  if(send(sfd, &ping_send, 1, 0) == 1) {
    printf("main: Sent ping: %c\n", ping_send);
  }
  else {
    rc=1; CheckRC("Send failed.\n");
  }

  if(recv(sfd, &ping_recv, 1, 0) == 1) {
    printf("main: Recv ping: %c\n", ping_recv);
  }
  else {
    rc=1; CheckRC("Receive failed.\n");
  }

  /*--- Clean up. ---*/
  comm_disengage_server(sfd);

  return(0);
}

#endif

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
