 /****************************************************************************/
 /*                                                                          */
 /*  Copyright (C) 1991-2000, all rights reserved                            */
 /*  Daniel Sleator, David Temperley, and John Lafferty                      */
 /*  See file "README" for information about commercial use of this system   */
 /*                                                                          */
 /****************************************************************************/
 /****************************************************************************/
 /*                                                                          */
 /*  Copyright (C) 2001-2003, additional rights reserve                      */
 /*  Daxtron Labs Inc, http://www.daxtron.com                                */
 /*  See file "README" for information about commercial use of this system   */
 /*   also "LGP-README.txt" for info about CMU restrictions                  */
 /*                                                                          */
 /****************************************************************************/

 /****************************************************************************
 *  
 *   This is a simple example of the link parser API.  It similates most of
 *   the functionality of the original link grammar parser, allowing sentences
 *   to be typed in either interactively or in "batch" mode (if -batch is
 *   specified on the command line, and stdin is redirected to a file).
 *   The program:
 *     Opens up a dictionary
 *     Iterates:
 *        1. Reads from stdin to get an input string to parse
 *        2. Tokenizes the string to form a Sentence
 *        3. Tries to parse it with cost 0
 *        4. Tries to parse with increasing cost
 *     When a parse is found:
 *        1. Extracts each Linkage
 *        2. Passes it to process_some_linkages()
 *        3. Deletes linkage
 *     After parsing each Sentence is deleted by making a call to 
 *     sentence_delete.
 *     
 ****************************************************************************/


#include "link-includes.h"
#include "command-line.h"
#ifndef WIN32
#define closesocket close
#else
#define close closesocket
#endif
#define systemError GetLastError()

#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <time.h>

#define MAXINPUT 1024
#define DISPLAY_MAX 32
#define COMMENT_CHAR '%'  /* input lines beginning with this are ignored */

static int batch_errors = 0;
static int input_pending=FALSE;
static Parse_Options  opts;
static Parse_Options  panic_parse_opts;
static char work_buf[32678];

int outputMode;
int MAX_DISPLAY;

short cyc_port;
char CycHost[128];

static char result_MT[256];
#define CYC_DEFAULT_SERVER_NAME "localhost"

#define SERVER_NAME "linkparser"
#define SERVER_URL "http://www.daxtron.com/linkparser/"
#define HTTPPROTOCOL "HTTP/1.0"
#define RFC1123FMT "%a, %d %b %Y %H:%M:%S GMT"

extern void initcache(void); // initalize the extrakb.txt processor
extern FILE *extrakbfile;

typedef enum {UNGRAMMATICAL='*', 
	      PARSE_WITH_DISJUNCT_COST_GT_0=':',
	      NO_LABEL=' '} Label;

int fget_input_string(char *input_string, FILE *in, FILE *out, 
		      Parse_Options opts) {
    if ((!parse_options_get_batch_mode(opts)) && (verbosity > 0)
	&& (!input_pending)) fprintf(out, "linkparser> ");
    fflush(out);
    input_pending = FALSE;
    if (fgets(input_string, MAXINPUT, in)) return 1;
    else return 0;
}

int fget_input_char(FILE *in, FILE *out, Parse_Options opts) {
    if (!parse_options_get_batch_mode(opts) && (verbosity > 0)) 
      fprintf(out, "linkparser> ");
    fflush(out);
    return getc(in);
}

/**************************************************************************
*  
*  This procedure displays a linkage graphically.  Since the diagrams 
*  are passed as character strings, they need to be deleted with a 
*  call to string_delete.
*
**************************************************************************/

void process_linkage(Linkage linkage, Parse_Options opts) {
    char * string;
    int    j, mode, first_sublinkage;

    if (parse_options_get_display_union(opts)) {
	linkage_compute_union(linkage);
	first_sublinkage = linkage_get_num_sublinkages(linkage)-1;
    }
    else {
	first_sublinkage = 0;
    }

    for (j=first_sublinkage; j<linkage_get_num_sublinkages(linkage); ++j) {
	linkage_set_current_sublinkage(linkage, j);
	if (parse_options_get_display_on(opts)) {
	    string = linkage_print_diagram(linkage);
	    fprintf(stdout, "%s", string);
	    string_delete(string);
	}
	if (parse_options_get_display_links(opts)) {
	    string = linkage_print_links_and_domains(linkage);
	    fprintf(stdout, "%s", string);
	    string_delete(string);
	}
	if (parse_options_get_display_postscript(opts)) {
	    string = linkage_print_postscript(linkage, FALSE);
	    fprintf(stdout, "%s\n", string);
	    string_delete(string);
	}
    }
    if ((mode=parse_options_get_display_constituents(opts))) {
	string = linkage_print_constituent_tree(linkage, mode);
	if (string != NULL) {
	    fprintf(stdout, "%s\n", string);
	    string_delete(string);
	} else {
	    fprintf(stderr, "Can't generate constituents.\n");
	    fprintf(stderr, "Constituent processing has been turned off.\n");
	}
    }
}

void print_parse_statistics(Sentence sent, Parse_Options opts) {
    if (sentence_num_linkages_found(sent) > 0) {
	if (sentence_num_linkages_found(sent) > 
	    parse_options_get_linkage_limit(opts)) {
	    fprintf(stdout, "Found %d linkage%s (%d of %d random " \
		    "linkages had no P.P. violations)", 
		    sentence_num_linkages_found(sent), 
		    sentence_num_linkages_found(sent) == 1 ? "" : "s",
		    sentence_num_valid_linkages(sent),
		    sentence_num_linkages_post_processed(sent));
	}
	else {
	    fprintf(stdout, "Found %d linkage%s (%d had no P.P. violations)", 
		    sentence_num_linkages_post_processed(sent), 
		    sentence_num_linkages_found(sent) == 1 ? "" : "s",
		    sentence_num_valid_linkages(sent));
	}
	if (sentence_null_count(sent) > 0) {
	    fprintf(stdout, " at null count %d", sentence_null_count(sent));
	}
	fprintf(stdout, "\n");
    }
}


void process_some_linkages(Sentence sent, Parse_Options opts) {
    int i,num_displayed, num_to_query;
    Linkage linkage;
   
    if (verbosity > 0) print_parse_statistics(sent, opts);
    if (!parse_options_get_display_bad(opts)) {
	num_to_query = MIN(sentence_num_valid_linkages(sent), MAX_DISPLAY);
    }
    else {
	num_to_query = MIN(sentence_num_linkages_post_processed(sent), 
			   MAX_DISPLAY);
    }

    for (i=0, num_displayed=0; i<num_to_query; ++i) {

	if ((sentence_num_violations(sent, i) > 0) &&
	    (!parse_options_get_display_bad(opts))) {
	    continue;
	}

	linkage = linkage_create(i, sent, opts);

	if (verbosity > 0) {
	  if ((sentence_num_valid_linkages(sent) == 1) &&
	      (!parse_options_get_display_bad(opts))) {
	    fprintf(stdout, "  Unique linkage, ");
	  }
	  else if ((parse_options_get_display_bad(opts)) &&
		   (sentence_num_violations(sent, i) > 0)) {
	    fprintf(stdout, "  Linkage %d (bad), ", i+1);
	  }
	  else {
	    fprintf(stdout, "  Linkage %d, ", i+1);
	  }
	  
	  if (!linkage_is_canonical(linkage)) {
	    fprintf(stdout, "non-canonical, ");
	  }
	  if (linkage_is_improper(linkage)) {
	    fprintf(stdout, "improper fat linkage, ");
	  }
	  if (linkage_has_inconsistent_domains(linkage)) {
	    fprintf(stdout, "inconsistent domains, ");
	  }
	  
	  fprintf(stdout, "cost vector = (UNUSED=%d DIS=%d AND=%d LEN=%d)\n",
		  linkage_unused_word_cost(linkage),
		  linkage_disjunct_cost(linkage),
		  linkage_and_cost(linkage),
		  linkage_link_cost(linkage));
	}

	process_linkage(linkage, opts);
	linkage_delete(linkage);

	//if (++num_displayed < num_to_query) {
	//    if (verbosity > 0) {
	//        fprintf(stdout, "Press RETURN for the next linkage.\n");
	//    }
	//    if ((c=fget_input_char(stdin, stdout, opts)) != '\n') {
	//	ungetc(c, stdin);
	//	input_pending = TRUE;
	//	break;
	//    }
	//}
    }
}

int there_was_an_error(Label label, Sentence sent, Parse_Options opts) {

    if (sentence_num_valid_linkages(sent) > 0) {
	if (label == UNGRAMMATICAL) {
	    batch_errors++;
	    return UNGRAMMATICAL;
	}
	if ((sentence_disjunct_cost(sent, 0) == 0) && 
	    (label == PARSE_WITH_DISJUNCT_COST_GT_0)) {
	    batch_errors++;
	    return PARSE_WITH_DISJUNCT_COST_GT_0;
	}
    } else {
	if  (label != UNGRAMMATICAL) {
	    batch_errors++;
	    return UNGRAMMATICAL;
	}
    }
    return FALSE;
}

void batch_process_some_linkages(Label label, Sentence sent, Parse_Options opts) {
    Linkage linkage;
   
    if (there_was_an_error(label, sent, opts)) {
	if (sentence_num_linkages_found(sent) > 0) {
	    linkage = linkage_create(0, sent, opts);
	    process_linkage(linkage, opts);
	    linkage_delete(linkage);
	}
	fprintf(stdout, "+++++ error %d\n", batch_errors);
    }
}

int special_command(char *input_string, Dictionary dict) {

    if (input_string[0] == '\n') return TRUE;
    if (input_string[0] == COMMENT_CHAR) return TRUE;
    if (input_string[0] == '!') {
        if (strncmp(input_string, "!panic_", 7)==0) {
	    issue_special_command(input_string+7, panic_parse_opts, dict);
	}
	else {
	    issue_special_command(input_string+1, opts, dict);
	}
	return TRUE;
    }
    return FALSE;
}

Label strip_off_label(char * input_string) {
    int c;

    c = input_string[0];
    switch(c) {
    case UNGRAMMATICAL:
    case PARSE_WITH_DISJUNCT_COST_GT_0:
	input_string[0] = ' ';
	return (Label)c;
	break;
    default:
	return NO_LABEL;
    }	
}

void setup_panic_parse_options(Parse_Options opts) {
    parse_options_set_disjunct_cost(opts, 3);
    parse_options_set_min_null_count(opts, 1);
    parse_options_set_max_null_count(opts, MAX_SENTENCE);
    parse_options_set_max_parse_time(opts, 60);
    parse_options_set_islands_ok(opts, 1);
    parse_options_set_short_length(opts, 6);
    parse_options_set_all_short_connectors(opts, 1);
    parse_options_set_linkage_limit(opts, 100);
}

void print_usage(char *str) {
    fprintf(stderr, 
	    "Usage: %s [dict_file] [-pp pp_knowledge_file]\n"
	    "          [-c constituent_knowledge_file] [-a affix_file]\n"
	    "          [-ppoff] [-coff] [-aoff] [-batch] [-<special \"!\" command>]\n", str);
    exit(-1);
}
// Modified to act as a TCP server process
//----------------------------------------
/*---------------------------------------------------------------------------
 * 
 * Synopsis
 * --------
 * Main program for server.
 *
 * Author
 * ------
 * Raymond McKendall
 * November 1998
 *
 *--------------------------------------------------------------------------*/

#include <signal.h>
#include "comm.h"

/*---------------------------------------------------------------------------*/

int start_service(
   u_short port,   /* Port for service.   */
   SOCKET* sfd     /* Socket for service. */
)
/*
 * Description
 * -----------
 * Creates socket. 
 * Binds socket to port.
 * Starts listening for connections to socket/port.
 *
 *---------------------------------------------------------------------------*/
{
  int rc;
  SocketAddressIN address;    /* Scocket address info for server. */

  /*--- Create socket. ---*/
  *sfd = socket(AF_INET, SOCK_STREAM, PROTOCOL);
  if (*sfd == SOCKET_ERROR) {
    comm_error("start_service: Cannot create socket");
    return(1);
  }

  /*--- Fill in socket address. ---*/
  memset(&address, 0, sizeof(SocketAddressIN));
  address.sin_family      = AF_INET;
  address.sin_port        = htons(port);
  address.sin_addr.s_addr = htonl(INADDR_ANY);

  /*--- Bind to port. ---*/
  rc= bind(*sfd, (SocketAddress *)&address, sizeof(SocketAddress));
  if (rc == SOCKET_ERROR) {
    comm_error("start_service: Cannot bind to port");
    return(1);
  }

  /*--- Listen for connections. ---*/
  if (listen(*sfd, 0) == SOCKET_ERROR) {
    comm_error("start_service: Cannot listen to port.");
    return(1);
  }

  return(0);
}

/*---------------------------------------------------------------------------*/

int accept_client(
   SOCKET  listen_socket,    /* Socket listened to for new connections. */
   SOCKET* active_socket,    /* Active socket connected to a client.    */
   HostEntry** client_info   /* Client accepted.                        */
)
/*
 * Description
 * -----------
 * Accepts connection from a client on listen_socket
 * and returns connection as active_socket and host entry.
 *
 *---------------------------------------------------------------------------*/
{
  SocketAddressIN client;
  int asize = sizeof(SocketAddressIN);

  /*--- Accept next connection ---*/
  *active_socket = accept(listen_socket, (SocketAddress *)&client, &asize);
  if (*active_socket == SOCKET_ERROR) {
    comm_error("accept_client: Failed to accept client connection");
    return(1);
  }
  
  *client_info = gethostbyname((char *) inet_ntoa(client.sin_addr));

  return(0);
}

/*---------------------------------------------------------------------------*/

int render_service(
   SOCKET listen_socket   /* Socket listened to for service requests. */
)
/*
 * Description
 * -----------
 * Accepts connection from a client on listen_socket and sends message.
 *
 *---------------------------------------------------------------------------*/
{
  char message[] = "?Hello, World!";
  SOCKET active_socket;     /* Socket for connection to client.    */
  HostEntry* client_info;   /* On active_socket.                   */
  char prompt;              /* Prompt from client to send message. */
  int rc;

  /*--- Wait for a client connection. ---*/
  rc= accept_client(listen_socket, &active_socket, &client_info);
  if (rc) {
    printf( "render_service: Failed accepting client.\n");
    return(1);
  }
 printf("render_service:Wait for a client connection\n");
  
  /*--- Receive prompt. ---*/
  rc= comm_recv_chars(active_socket, &prompt, 2);
  if(rc) {
    printf( "render_service: Failed receiving prompt.\n");
    return(1);
  }
  printf("render_service:Receive prompt\n");
  /*--- Send message. ---*/
  message[0] = (char) strlen(message);
  rc= comm_send_chars(active_socket, message, message[0]+1);
  if (rc) {
    printf( "render_service: Failed sending message.\n");
    return(1);
  }
  printf("render_service:message sent\n");
 
  /*--- Finish. ---*/
  printf("*server: Sent message \"%s\" \n", message+1);
 // printf(" to client \"%s\".\n", client_info->h_name);

  printf("render_service:message sent\n");
  comm_close_socket(active_socket);
  printf("render_service:message sent\n");

  return(0);
}

/*---------------------------------------------------------------------------*/
/*--- main() ----------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
int set_blocking(SOCKET sfd, int block);
void inferlinks(SOCKET sock, Linkage linkage);
void cyc_inferlinks(SOCKET sock, Linkage linkage);
void cyc_inferlinks_cycl(SOCKET sock, Linkage linkage,char *);
void load_cdrules(void);

void sock_print(SOCKET sock,char *message)
{
	int size;
	int totalSent;
	int bytesSent;
		  size= strlen(message);
		  //int totalSent;
		  totalSent= 0;
		  do {
			 bytesSent = send(sock, message + totalSent,
								 strlen(message + totalSent), 0);
			if (bytesSent == SOCKET_ERROR)
			{
			  printf("THROW send");
			  break;
			}
			totalSent += bytesSent;
		  } while (totalSent < size);
//	if (outputMode == 1) //HTML
//	{
//	 bytesSent = send(sock, "<BR>\n",5, 0);
//	}
}

#define CheckRC(msg) {if(rc) {fprintf(stderr, "%s\n", msg); exit(1);}}

static void
send_headers(SOCKET sock, int status, char* title, char* extra_header, char* mime_type, long length, time_t mod )
    {
    time_t now;
    char timebuf[100];

    sprintf(work_buf, "%s %d %s\r\n", HTTPPROTOCOL, status, title ); sock_print(sock,work_buf);
    sprintf(work_buf,"Server: %s\r\n", SERVER_NAME );sock_print(sock,work_buf);
    now = time( (time_t*) 0 );
    (void) strftime( timebuf, sizeof(timebuf), RFC1123FMT, gmtime( &now ) );
    sprintf(work_buf,"Date: %s\r\n", timebuf );sock_print(sock,work_buf);
    if ( extra_header != (char*) 0 )
	sprintf(work_buf, "%s\r\n", extra_header );sock_print(sock,work_buf);
    if ( mime_type != (char*) 0 )
	sprintf(work_buf,"Content-Type: %s\r\n", mime_type );sock_print(sock,work_buf);
    if ( length >= 0 )
	sprintf(work_buf,"Content-Length: %ld\r\n", (long) length );sock_print(sock,work_buf);
    if ( mod != (time_t) -1 )
	{
	(void) strftime( timebuf, sizeof(timebuf), RFC1123FMT, gmtime( &mod ) );
	sprintf(work_buf,"Last-Modified: %s\r\n", timebuf ); sock_print(sock,work_buf);
	}
    sprintf(work_buf,"Connection: close\r\n" );sock_print(sock,work_buf);
    sprintf(work_buf, "\r\n" );sock_print(sock,work_buf);
    }

static void
send_error( SOCKET sock, int status, char* title, char* extra_header, char* text )
    {
    send_headers( sock, status, title, extra_header, "text/html", -1, -1 ); sock_print(sock,work_buf);
    sprintf(work_buf, "<HTML><HEAD><TITLE>%d %s</TITLE></HEAD>\n<BODY BGCOLOR=\"#cc9999\"><H4>%d %s</H4>\n", status, title, status, title ); sock_print(sock,work_buf);
    sprintf(work_buf, "%s\n", text ); sock_print(sock,work_buf);
    sprintf(work_buf,"<HR>\n<ADDRESS><A HREF=\"%s\">%s</A></ADDRESS>\n</BODY></HTML>\n", SERVER_URL, SERVER_NAME ); sock_print(sock,work_buf);
    }

static char* get_mime_type( char* name )
    {
    char* dot;

    dot = strrchr( name, '.' );
    if ( dot == (char*) 0 )
	return "text/plain; charset=iso-8859-1";
    if ( strcmp( dot, ".html" ) == 0 || strcmp( dot, ".htm" ) == 0 )
	return "text/html; charset=iso-8859-1";
    if ( strcmp( dot, ".jpg" ) == 0 || strcmp( dot, ".jpeg" ) == 0 )
	return "image/jpeg";
    if ( strcmp( dot, ".gif" ) == 0 )
	return "image/gif";
    if ( strcmp( dot, ".png" ) == 0 )
	return "image/png";
    if ( strcmp( dot, ".css" ) == 0 )
	return "text/css";
    if ( strcmp( dot, ".au" ) == 0 )
	return "audio/basic";
    if ( strcmp( dot, ".wav" ) == 0 )
	return "audio/wav";
    if ( strcmp( dot, ".avi" ) == 0 )
	return "video/x-msvideo";
    if ( strcmp( dot, ".mov" ) == 0 || strcmp( dot, ".qt" ) == 0 )
	return "video/quicktime";
    if ( strcmp( dot, ".mpeg" ) == 0 || strcmp( dot, ".mpe" ) == 0 )
	return "video/mpeg";
    if ( strcmp( dot, ".vrml" ) == 0 || strcmp( dot, ".wrl" ) == 0 )
	return "model/vrml";
    if ( strcmp( dot, ".midi" ) == 0 || strcmp( dot, ".mid" ) == 0 )
	return "audio/midi";
    if ( strcmp( dot, ".mp3" ) == 0 )
	return "audio/mpeg";
    if ( strcmp( dot, ".pac" ) == 0 )
	return "application/x-ns-proxy-autoconfig";
    return "text/plain; charset=iso-8859-1";
    }

/* Signal handling. */
int active = 1;
void signal_handler (int signum){active = 0;}

int serv_main(u_short port)
{
  SOCKET listen_socket;
//  u_short port;
  int rc;

 
  /*--- Initialize socket support ---*/
  rc= comm_initialize_sockets();
  CheckRC("*client: Could not initialize socket support.\n");

  /*--- Start service. ---*/
  rc= start_service(port, &listen_socket);
  CheckRC("*server: Could not start service.");
  printf("*server: Listening to port %d.\n", port);
  /*--- Toggle blocking. ---*/
  rc= set_blocking(listen_socket, 0);   CheckRC("Failed to set non-blocking mode.\n");
 // rc= set_blocking(listen_socket, 1);   CheckRC("Failed to set blocking mode.\n");

  /*--- Signal handling. ---*/
  signal(SIGTERM, signal_handler);
  signal(SIGINT,  signal_handler);

  /*--- Render service. ---*/
  while(active) {
	printf("render_service:begin\n");
    render_service(listen_socket);
	printf("render_service:fin\n");
  }

  /*--- Done. ---*/
  printf("*server: Done.\n");
  comm_close_socket(listen_socket);
  comm_terminate_sockets();
  return(0);
}

// Utilities

static char buf1[512], buf2[512];

int indexOf(char *str, char *target)
{
  char *p = strstr(str, target);
  if (p == 0)  return -1;
  return (int)p - (int)str + 1;
}


int indexOfIgnoreCase(char *str, char *target)
{
  int i, len = strlen(str); if (len > 510) len = 510;
  for (i=0; i<len; i++) buf1[i] = tolower(str[i]);
  buf1[len] = 0;
  len = strlen(target);
  for (i=0; i<len; i++) buf2[i] = tolower(target[i]);
  buf2[len] = 0;
  return indexOf(buf1, buf2);
}

void toLower(char *str)
{
  int i, len = strlen(str);
  for (i=0; i<len; i++) str[i] = tolower(str[i]);
}

void decodeEncode(char *s) {
  static char buf[350];
  int count;
  int len;
  int i;
  count=0;
  len= strlen(s);
  for (i=0; i<len; i++) {
	  if ((s[i]=='%')&&(s[i+1]=='2')&&(s[i+2]>='0'))
	  {s[count++]=0x20+(s[i+2]-'0');
	   i=i+2;
	  }
	else
	{
		s[count++]=s[i];
	}
  }
  s[count++] = 0;
}


void writeOutputHeader(SOCKET sock,char * title)
{
  int count;
  count = 0; 
  work_buf[0] = 0;
  if (outputMode == 1) {  // HTML
    count = sprintf((char *)work_buf, "<HTML><TITLE>%s</TITLE>\n<H2>%s</H2>\n  <BODY>\n", title, title);
	sock_print(sock,work_buf);  
  } else if ((outputMode == 2)||(outputMode == 4)) { // XML
    count= sprintf((char *)work_buf, "<?xml version=\"1.0\" ?>\n", title);
	sock_print(sock,work_buf);
    count= sprintf((char *)work_buf, "<!-- %s -->\n", title);
	sock_print(sock,work_buf);
//    count += sprintf((char *)work_buf, "<CDdata>\n");
//	sock_print(sock,work_buf);
  } else {  // TEXT or unknown
  }

}
void writeOutputTrailer(SOCKET sock)
{
  int count;
  if (outputMode == 1) {  // HTML
    count = sprintf((char *)work_buf, "  </BODY>\n</HTML>\n");
	sock_print(sock,work_buf);
  } else if ((outputMode == 2)||(outputMode == 4)) { // XML
//    count += sprintf((char *)work_buf, "</CDdata>\n");
//	sock_print(sock,work_buf);
  } else {  // TEXT or unknown
  }

}

//----------------------------------------
#define backlog 20
#define length 8196

  char recvBuffer[length + 1];   // Null terminated
  char message[length+1];
  char lbuf[MAXINPUT+1];
  char rbuf[MAXINPUT+1];

int main(int argc, char * argv[]) {

    Dictionary      dict;
    Sentence        sent;
    char            *dictionary_file=NULL;
    char            *post_process_knowledge_file=NULL;
    char            *constituent_knowledge_file=NULL;
    char            *affix_file=NULL;
    int             pp_on=TRUE;
    int             af_on=TRUE;
    int             cons_on=TRUE;
    int             num_linkages, i;
    char            input_string[MAXINPUT];
    Label           label = NO_LABEL;  
    int             parsing_space_leaked, reported_leak, dictionary_and_option_space;

//--- SERVER VARS ----
//  SOCKET listen_socket;
  u_short port;
  int rc;
  //char message[256];
//  SOCKET active_socket;     /* Socket for connection to client.    */
//  HostEntry* client_info;   /* On active_socket.                   */
  //char prompt[MAXINPUT];              /* Prompt from client to send message. */
 // int rc;
  Linkage       linkage;
  char *        diagram;
  char *        auxword;
  int	w,l,r;
  int N_links;
  int link;
  char *klabel, *llabel, *rlabel;

	char *lp;
	char *rp;

 // int i,  
 int num_displayed, num_to_query;
 int count_to;
 int index, index2;
 int index3;
 FILE *flog;
 //int size;
 //int totalSent;
 int total_valid_linkages;
 int totalReceived;
 struct hostent* entity;
 //SOCKET serverSocket;
 struct sockaddr_in serverAddr;
struct sockaddr_in cyc_serverAddr;


  SOCKET serverSocket = INVALID_SOCKET;
  SOCKET sock         = INVALID_SOCKET;
  SOCKET cyc_sock   = INVALID_SOCKET;
  struct hostent *hp;
  char *lmsg[256];

  port=2002;
  cyc_port = 3601; // default cyc port
  strcpy(CycHost,CYC_DEFAULT_SERVER_NAME);

	load_cdrules();

    i = 1;
    if ((argc > 1) && (argv[1][0] != '-')) {
	/* the dictionary is the first argument if it doesn't begin with "-" */
	dictionary_file = argv[1];	
	i++;
    }

    for (; i<argc; i++) {
	if (argv[i][0] == '-') {
	    if (strcmp("-pp", argv[i])==0) {
		if ((post_process_knowledge_file != NULL) || (i+1 == argc)) 
		  print_usage(argv[0]);
		post_process_knowledge_file = argv[i+1];
		i++;
	    } else 
	    if (strcmp("-c", argv[i])==0) {
		if ((constituent_knowledge_file != NULL) || (i+1 == argc)) 
		  print_usage(argv[0]);
		constituent_knowledge_file = argv[i+1];
		i++;
	    } else 
	    if (strcmp("-a", argv[i])==0) {
		if ((affix_file != NULL) || (i+1 == argc)) print_usage(argv[0]);
		affix_file = argv[i+1];
		i++;
	    } else if (strcmp("-ppoff", argv[i])==0) {
		pp_on = FALSE;
	    } else if (strcmp("-coff", argv[i])==0) {
		cons_on = FALSE;
	    } else if (strcmp("-aoff", argv[i])==0) {
		af_on = FALSE;
	    } else if (strcmp("-batch", argv[i])==0) {
	    } else if (strncmp("-!", argv[i],2)==0) {
	    } else {
		print_usage(argv[0]);		
	    }
	} else {
	    print_usage(argv[0]);
	}
    }

    if (!pp_on && post_process_knowledge_file != NULL) print_usage(argv[0]);

    if (dictionary_file == NULL) {
	dictionary_file = "4.0.dict";
        fprintf(stderr, "No dictionary file specified.  Using %s.\n", 
		dictionary_file);
    }

    if (af_on && affix_file == NULL) {
	affix_file = "4.0.affix";
        fprintf(stderr, "No affix file specified.  Using %s.\n", affix_file);
    }

    if (pp_on && post_process_knowledge_file == NULL) {
	post_process_knowledge_file = "4.0.knowledge";
        fprintf(stderr, "No post process knowledge file specified.  Using %s.\n",
		post_process_knowledge_file);
    }

    if (cons_on && constituent_knowledge_file == NULL) {
        constituent_knowledge_file = "4.0.constituent-knowledge"; 
	fprintf(stderr, "No constituent knowledge file specified.  Using %s.\n", 
		constituent_knowledge_file);
    }

    opts = parse_options_create();
    if (opts == NULL) {
	fprintf(stderr, "%s\n", lperrmsg);
	exit(-1);
    }

    panic_parse_opts = parse_options_create();
    if (panic_parse_opts == NULL) {
	fprintf(stderr, "%s\n", lperrmsg);
	exit(-1);
    }
    setup_panic_parse_options(panic_parse_opts);
    parse_options_set_max_sentence_length(opts, 70);
    parse_options_set_panic_mode(opts, TRUE);
    parse_options_set_max_parse_time(opts, 30);
    parse_options_set_linkage_limit(opts, 1000);
    parse_options_set_short_length(opts, 10);

    dict = dictionary_create(dictionary_file, 
			     post_process_knowledge_file,
			     constituent_knowledge_file,
			     affix_file);
    if (dict == NULL) {
	fprintf(stderr, "%s\n", lperrmsg);
	exit(-1);
    }

    /* process the command line like commands */
    for (i=1; i<argc; i++) {
	if ((strcmp("-pp", argv[i])==0) || 
	    (strcmp("-c", argv[i])==0) || 
	    (strcmp("-a", argv[i])==0)) {
	  i++;
	} else if ((argv[i][0] == '-') && (strcmp("-ppoff", argv[i])!=0) &&
		   (argv[i][0] == '-') && (strcmp("-coff", argv[i])!=0) &&
		   (argv[i][0] == '-') && (strcmp("-aoff", argv[i])!=0)) {
	  issue_special_command(argv[i]+1, opts, dict);
	}
    }

    dictionary_and_option_space = space_in_use;  
    reported_leak = external_space_in_use = 0;
    verbosity = parse_options_get_verbosity(opts);
	initcache();

    //AfxBeginThread(ServerThread,0);
    //serv_main(2002); // <--- EXTERNAL SERVICE
if(2==2)
{

    //int rc = 0;
#ifdef WIN32
    WSADATA wsaData;
    rc = WSAStartup(MAKEWORD(1, 1), &wsaData);
    if (rc != 0)
	{
      printf( "THROW Cannot find a usable Windows socket library.\n");
	  return -1;//break;
	}
    if (LOBYTE(wsaData.wVersion) != 1 || HIBYTE(wsaData.wVersion) != 1) {
      WSACleanup();
      printf( "THROW Cannot find a usable Windows socket library.\n");
	  return -1; //break;
    }
#endif
	serverSocket	= socket(AF_INET, SOCK_STREAM, 0);
    if (serverSocket == INVALID_SOCKET)
	{
      printf( "THROW socket\n");
	  return -1; //break;
	}
     serverAddr.sin_family      = AF_INET;
    serverAddr.sin_port        = htons(port);
    serverAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    memset(&(serverAddr.sin_zero), 0, 8);
    rc = bind(serverSocket, (struct sockaddr*) &serverAddr,
              sizeof(struct sockaddr));
    if (rc == -1)
	{
      printf( "THROW bind\n");
	  return -1; //break;
	}
    rc = listen(serverSocket, backlog);
    if (rc == -1)
	{
      printf( "THROW listen\n");
	  return -1; //break;
	}
    printf( "Accepting connections ...\n" );


    while (1) {
      struct sockaddr_in clientAddr;
      int clientAddrSize = sizeof(struct sockaddr_in);
      do
        sock = accept(serverSocket,
                      (struct sockaddr*) &clientAddr, 
                      &clientAddrSize);
      while ((sock == INVALID_SOCKET) && (errno == EINTR));
      if (sock == INVALID_SOCKET)
	  {
        printf( "THROW accept\n");
		continue;
	  }
      //struct hostent* 
	  entity = NULL;
      entity = gethostbyaddr((const char*) &clientAddr.sin_addr,
                             sizeof(struct in_addr), AF_INET);
      //if (DEBUG) printf("Connection from "); 
      //                << inet_ntoa((struct in_addr) clientAddr.sin_addr)
      //                << " (" << ((entity == NULL) ? "error" : entity->h_name)
      //               << ")." << endl;
      //int totalReceived;
	   printf("Connection from %s ",(char *)inet_ntoa(clientAddr.sin_addr));
	  if (entity!=NULL) printf("(%s)",entity->h_name );
	  printf("\n");
      do {
        totalReceived = 0;
        do {
          int bytesReceived; 
			  bytesReceived= recv(sock, recvBuffer + totalReceived,
                                   length - totalReceived, 0);
          if (bytesReceived == SOCKET_ERROR)
		  {
            printf( "THROW recv\n");
			break;
		  }
          totalReceived += bytesReceived;
          //if (DEBUG) cout << "bytes received=" << bytesReceived << "\n";
        } while ((totalReceived > 0 )&& 
                 (totalReceived < length) && 
                 (recvBuffer[totalReceived - 1] != '\n') );

		extrakbfile=fopen("extrakb.txt","r"); // Open up the lexical xml data cache

        if (totalReceived > 0) {
          if (recvBuffer[totalReceived - 1] != '\n')
		  {
		  
			send_error(sock, 400, "Bad Request", (char*) 0, "Receive buffer overflow. " );

			break;
		  }
          recvBuffer[totalReceived] = '\0';
          //if (DEBUG) {
          // cout << "Received from client: " << recvBuffer;
          //  cout.flush();
          // }
		  //------------------------------------------------
			// First, figure out what type of output was requested:
		  MAX_DISPLAY=DISPLAY_MAX;
			outputMode = 3; // default to TEXT?  // yes, this is a global variable defined in nlpserver.h
			index = indexOfIgnoreCase(recvBuffer, "html?");
			if (index > -1)  outputMode = 1;
			index = indexOfIgnoreCase(recvBuffer, "xml?");
			if (index > -1)  outputMode = 2;
			index = indexOfIgnoreCase(recvBuffer, "cdml?");
			if (index > -1)  outputMode = 4;
			index = indexOfIgnoreCase(recvBuffer, "cdml1?");
			if (index > -1) { outputMode = 4; MAX_DISPLAY=1;}


			decodeEncode(recvBuffer);

				// check for a Target Microtheory request:
			    result_MT[0]=0;
				index = indexOfIgnoreCase(recvBuffer, "defaultmt=");
				if (index > -1) {
				  index2 = index + 10;
				  index3= -1;
				  for (i=index2+1; i<(int)strlen(recvBuffer); i++) {
					if (recvBuffer[i] == '"') {
					  index3 = i;
					  break;
					}
				  }
				  if (index3 > -1) {
					int count;
					int i;
					count = 0;
					for ( i=index2; i<index3; i++) result_MT[count++] = recvBuffer[i];
					result_MT[count] = 0;
					printf("Default_MT = [%s]\n",result_MT);
					decodeEncode(recvBuffer);
				  }
				}

				// check for a Part Of Speech request:
				index = indexOfIgnoreCase(recvBuffer, "pos=");
				if (index > -1) {
				  index2 = index + 4;
				  index3= -1;
				  for (i=index2+1; i<(int)strlen(recvBuffer); i++) {
					if (recvBuffer[i] == '"') {
					  index3 = i;
					  break;
					}
				  }
				  if (index3 > -1) {
					int count;
					int i;
					count = 0;
					for ( i=index2; i<index3; i++) recvBuffer[count++] = recvBuffer[i];
					recvBuffer[count] = 0;
					printf("POS request:=|%s|, index2=%d, index3=%d\n", recvBuffer, index2, index3);
					decodeEncode(recvBuffer);
				  }
				}
				// check for a botched HTTP REQUEST:
				index = indexOfIgnoreCase(recvBuffer, "HTTP");
				if (index>-1)
				{
					// pretend we heard nothing
					if (outputMode==1)
					{
					send_error(sock, 400, "Bad Request", (char*) 0, "Can't parse request. 'pos=' may be missing. Use http://server:2002/HTML?pos=\" this is a test. \"" );
					recvBuffer[0]='.';
					recvBuffer[1]=0;
					}
					else
					{
						writeOutputHeader(sock,"linkanalysis");
						sock_print(sock,"<linkgram>\r\n");
						sock_print(sock,"<origsent></origsent>\r\n");
						sock_print(sock,"<wordcount>0</wordcount>\r\n");
						sock_print(sock,"<parsecount>0</parsecount>\r\n");
						sock_print(sock,"<totalparses>0</totalparses>\r\n");
						sock_print(sock,"<rawwords></rawwords>\r\n");
						sock_print(sock,"</linkgram>\r\n");

					}
					break;
				}
				if (outputMode==1) 
					send_headers( sock,200, "Ok", (char*) 0, get_mime_type( "parse.html" ),-1,-1);// sb.st_size, sb.st_mtime );
				else
					writeOutputHeader(sock,"linkanalysis");

		  		  /*--- PROCESS MESSAGE HERE --*/
				if (space_in_use != dictionary_and_option_space + reported_leak) {
					fprintf(stderr, "Warning: %d bytes of space leaked.\n",
						space_in_use-dictionary_and_option_space-reported_leak);
					reported_leak = space_in_use - dictionary_and_option_space;
				}

				if ((strcmp(recvBuffer, "quit\n")==0) ||
					(strcmp(recvBuffer, "exit\n")==0)) 
				{
					active=0;
					break;
				}
				if (special_command(recvBuffer, dict)) continue;
				if (parse_options_get_echo_on(opts)) {
					printf("%s", input_string);
				}
				printf("input =[%s]\n", recvBuffer);
				if (parse_options_get_batch_mode(opts)) {
					label = strip_off_label(recvBuffer);
				}

				sent = sentence_create(recvBuffer, dict);

				if (sent == NULL) {
					if (verbosity > 0) fprintf(stderr, "%s\n", lperrmsg);
					//if (lperrno != NOTINDICT) exit(-1);
					if (lperrno != NOTINDICT) 
					{
						writeOutputHeader(sock,"linkanalysis");
						sock_print(sock,"<linkgram>\r\n");
						sock_print(sock,"<origsent></origsent>\r\n");
						sock_print(sock,"<wordcount>0</wordcount>\r\n");
						sock_print(sock,"<parsecount>0</parsecount>\r\n");
						sock_print(sock,"<totalparses>0</totalparses>\r\n");
						sock_print(sock,"<rawwords></rawwords>\r\n");
						sock_print(sock,"</linkgram>\r\n");
						break;
					}
					else continue;
				} 
				if (sentence_length(sent) > parse_options_get_max_sentence_length(opts)) {
					sentence_delete(sent);
					if (verbosity > 0) {
					  fprintf(stdout, 
						  "Sentence length (%d words) exceeds maximum allowable (%d words)\n",
						sentence_length(sent), parse_options_get_max_sentence_length(opts));
					}
					continue;
				}

				/* First parse with cost 0 or 1 and no null links */
				parse_options_set_disjunct_cost(opts, 2);
				parse_options_set_min_null_count(opts, 0);
				parse_options_set_max_null_count(opts, 0);
				parse_options_reset_resources(opts);

				num_linkages = sentence_parse(sent, opts);

				/* Now parse with null links */
				if ((num_linkages == 0) && (!parse_options_get_batch_mode(opts))) {
					if (verbosity > 0) fprintf(stdout, "No complete linkages found.\n");
					if (parse_options_get_allow_null(opts)) {
					parse_options_set_min_null_count(opts, 1);
					parse_options_set_max_null_count(opts, sentence_length(sent));
					num_linkages = sentence_parse(sent, opts);
					}
				}

				if (parse_options_timer_expired(opts)) {
					if (verbosity > 0) fprintf(stdout, "Timer is expired!\n");
				}
				if (parse_options_memory_exhausted(opts)) {
					if (verbosity > 0) fprintf(stdout, "Memory is exhausted!\n");
				}

				if ((num_linkages == 0) && 
					parse_options_resources_exhausted(opts) &&
					parse_options_get_panic_mode(opts)) {
					print_total_time(opts);
					if (verbosity > 0) fprintf(stdout, "Entering \"panic\" mode...\n");
					parse_options_reset_resources(panic_parse_opts);
					parse_options_set_verbosity(panic_parse_opts, verbosity);
					num_linkages = sentence_parse(sent, panic_parse_opts);
					if (parse_options_timer_expired(panic_parse_opts)) {
					if (verbosity > 0) fprintf(stdout, "Timer is expired!\n");
					}
				}

				print_total_time(opts);

			//	if (parse_options_get_batch_mode(opts)) {
			//		batch_process_some_linkages(label, sent, opts);
			//	}
			//	else {
			//		process_some_linkages(sent, opts);
			//	}

			 if (!parse_options_get_display_bad(opts)) {
			num_to_query = MIN(sentence_num_valid_linkages(sent), MAX_DISPLAY);
			}
			else {
			num_to_query = MIN(sentence_num_linkages_post_processed(sent), 
					   MAX_DISPLAY);
			}
					    /*--- Toggle blocking. ---*/

		  //rc= set_blocking(active_socket, 1);   CheckRC("Failed to set blocking mode.\n");
			sprintf(message,"\r\n<linkgram>\r\n");
			sock_print(sock,message);

			if (outputMode==1) 
				sprintf(message,"original input ='%s'<BR>\r\n",recvBuffer);
			else
				sprintf(message,"<origsent>%s</origsent>\r\n",recvBuffer);
			sock_print(sock,message);

			

			count_to=sentence_length(sent);
			total_valid_linkages=sentence_num_valid_linkages(sent);
			if (outputMode==1)
				sprintf(message,"wordcount=%i <BR> parsecount=%i<BR>totalparses=%i<BR>",count_to,num_to_query,total_valid_linkages);
			else
				sprintf(message,"<wordcount>%i</wordcount>\r\n<parsecount>%i</parsecount>\r\n<totalparses>%i</totalparses>\r\n",count_to,num_to_query,total_valid_linkages);
			sock_print(sock,message);
 			
			if (outputMode==1)
				sprintf(message,"rawwords<BR>");
			else
				sprintf(message,"<rawwords>");
			sock_print(sock,message);

			if (outputMode==1) sock_print(sock,"<UL>\r\n");
			for(i=0;i<count_to;i++)
			{
				diagram = sentence_get_word(sent, i);
				if (outputMode==1)
					sprintf(message,"<LI>word%i=%s</LI>",i,diagram);
				else
					sprintf(message,"<word>%s</word>",diagram);
				sock_print(sock,message);
 			    //string_delete(diagram);

			}
			if (outputMode==1) sock_print(sock,"</UL>\r\n");

			if (outputMode!=1) 
			{
				sprintf(message,"</rawwords>\r\n");
				sock_print(sock,message);
			}

			for (i=0, num_displayed=0; i<num_to_query; ++i) {
   
				linkage = linkage_create(i, sent, opts);

				if (outputMode==1)
					sprintf(message,"<H2>interpertation %i</H2><BR> wordset<BR>\r\n",i);
				else
					sprintf(message,"\r\n<interpertation>\r\n<wordset>");
 			    sock_print(sock,message);

				count_to = linkage_get_num_words(linkage);
 				for(w=0;w<count_to;w++)
				{
					auxword = linkage_get_word(linkage, w);
					if (outputMode==1)
						sprintf(message,"<LI>%s</LI>\r\n",auxword);
					else
						sprintf(message,"<word>%s</word>",auxword);
 					sock_print(sock,message);
				}

				if (outputMode==1)
					sprintf(message,"</UL>\r\n");
				else
					sprintf(message,"</wordset>\r\n");

 			    sock_print(sock,message);

				N_links = linkage_get_num_links(linkage);
				
				if (outputMode==1)
					sprintf(message,"Linkcount=%i<BR>\r\n",N_links);
				else
					sprintf(message,"<linkcount>%i</linkcount>\r\n",N_links);
				sock_print(sock,message);

				if (outputMode==1) sock_print(sock,"<UL>\r\n");

				for(link=0;link<N_links;link++)
				{
				    l      = linkage_get_link_lword(linkage, link);
					r      = linkage_get_link_rword(linkage, link);
					klabel  = linkage_get_link_label(linkage, link);
					llabel = linkage_get_link_llabel(linkage, link);
					rlabel = linkage_get_link_rlabel(linkage, link);
					if (outputMode==1)
						sprintf(message,"   <LI>%s--[%s]--%s</LI>\r\n",
											linkage_get_word(linkage,l),klabel,linkage_get_word(linkage, r));
					else
						sprintf(message,"<link><llink>%s</llink><label>%s</label><rlink>%s</rlink></link>\r\n",
											linkage_get_word(linkage,l),klabel,linkage_get_word(linkage, r));
 						sock_print(sock,message);

				}

				if (outputMode==1) sock_print(sock,"</UL><BR>\r\n");
				// Note to the rule guessing files
				flog = fopen("cycR1.txt","a");
				if(flog!=NULL)
				{
					fprintf(flog,"\r\n<origsent>%s</origsent>\r\n",recvBuffer);
					fclose(flog);
				}

				flog = fopen("cycR2.txt","a");
				if(flog!=NULL)
				{
					fprintf(flog,"\r\n<origsent>%s</origsent>\r\n",recvBuffer);
					fclose(flog);
				}

				// Hypolinks
				if (outputMode==1) 
					sock_print(sock,"<UL>\r\n");
				else
					sock_print(sock,"<hypolinks>\r\n");
				inferlinks(sock, linkage);
				if (outputMode==1) 
					sock_print(sock,"</UL><BR>\r\n");
				else
					sock_print(sock,"</hypolinks>\r\n");
				// End Hypolinks


				// Cyc-Hypolinks
				if (outputMode==1) 
					sock_print(sock,"<UL>\r\n");
				else
				{ 
					//sock_print(sock,"<cyc-hypolinks>\r\n");
				    //sock_print(sock,"(#$thereExists ?SENT \r\n");
					//sock_print(sock,"  (#$isa ?SENT #$LGP-SENT)\r\n");
					//sock_print(sock,"    (and \r\n");
				}
				cyc_inferlinks(sock, linkage);

				//------ MOD TO SEND CYCL TO local OpenCyc Engine
				if (result_MT[0]!=0)
				{
					// We did get a defaultmt request
					cyc_sock = socket(AF_INET,SOCK_STREAM,0);

					if (cyc_sock == INVALID_SOCKET)
					{ 
						printf("BAD Cyc_Socket \n");
					}
					else
					{
					//	int nodelayflag=1;
					//	rc = setsockopt(cyc_sock,IPPROTO_TCP,TCP_NODELAY,&nodelayflag,sizeof(nodelayflag));
						cyc_serverAddr.sin_family      = AF_INET;
						cyc_serverAddr.sin_port        = htons(cyc_port);
						hp = gethostbyname(CycHost);
						if (hp==NULL)
						{
							printf("CycHost %s not found! \n",CycHost);

						}
						else
						{
							memset(&(cyc_serverAddr.sin_zero), 0, 8);
							memcpy(&(cyc_serverAddr.sin_addr),hp->h_addr_list[0] ,hp->h_length);
							
							//cyc_serverAddr.sin_addr.s_addr = htonl(inet_addr(gethostbyname(CycHost))); //htonl(INADDR_ANY);
							rc=connect(cyc_sock,(struct sockaddr*)&cyc_serverAddr,sizeof(struct sockaddr));
							if (rc==-1)
							{
								printf("Cyc_sock connection failed. \n");
							}
							else
							{
							printf("Asserting CycL to %s post= %i in MT=%s \n",CycHost,cyc_port, result_MT);
							
							//sock_print(cyc_sock,"()\n (CYC-ASSERT '");
							cyc_inferlinks_cycl(cyc_sock, linkage,result_MT);
							//sock_print(cyc_sock," #$");
							//sock_print(cyc_sock,result_MT);
							//sock_print(cyc_sock," ) \n");
							//sock_print(cyc_sock,"(api-quit)\n\n");
							rc=read(cyc_sock,lmsg,256);
							lmsg[255]=0;
							//printf("CycReply :%s\n",lmsg);

							close(cyc_sock);
							}
						}
					}
				}


				if (outputMode==1) 
					sock_print(sock,"</UL><BR>\r\n");
				else
				{
					//sock_print(sock,"  ))\r\n");
					//sock_print(sock,"</cyc-hypolinks>\r\n");
				}
				// End Cyc-Hypolinks
				if (outputMode!=1) 
					
				{

					// PROLOG LINK OUTPUT
					sock_print(sock,"<prologlinks>\r\n links([");

					for(link=0;link<N_links;link++)
					{
						l      = linkage_get_link_lword(linkage, link);
						r      = linkage_get_link_rword(linkage, link);
						klabel  = linkage_get_link_label(linkage, link);
						strcpy(lbuf,linkage_get_word(linkage,l));
						strcpy(rbuf,linkage_get_word(linkage,r));
						lp = strchr(lbuf,'.');
						rp = strchr(rbuf,'.');
						if (lp==NULL)
							{ 
								lp="-";
							}
							else
							{
								*lp=0;
								lp++;

							}
						if (rp==NULL)
							{ 
								rp="-";
							}
							else
							{
								*rp=0;
								rp++;

							}

						if (link<N_links-1)
							sprintf(message," l(('%s'),w(('%s'),('%s')),w(('%s'),('%s'))),",
												klabel,lbuf,lp,rbuf,rp);
						else
							sprintf(message,"  l(('%s'),w(('%s'),('%s')),w(('%s'),('%s')))",
												klabel,lbuf,lp,rbuf,rp);
 						toLower(message);	
						sock_print(sock,message);
						
					}

					sock_print(sock,"])\r\n</prologlinks>\r\n");
				}



				if (outputMode==1)
				{
					//HTML
					diagram = linkage_print_constituent_tree(linkage,1);
					//printf("%s\n",diagram);
					sprintf(message,"Parse '%s'<BR>\r\n",diagram);
				}
				else
				{
					//Everything Else
					diagram = linkage_print_constituent_tree(linkage,1);
					//printf("%s\n",diagram);
					sprintf(message,"%s\r\n</interpertation>\r\n",diagram);
				}
				sock_print(sock,message);
				
			  /*--- Finish. ---*/
			 //printf("*server: Sent message \"%s\" \n", message+1);
			 // printf(" to client \"%s\".\n", client_info->h_name);

			  string_delete(diagram);
			  linkage_delete(linkage);

			  if (external_space_in_use != 0) {
				fprintf(stderr, "Warning: %d bytes of external space leaked.\n", 
				external_space_in_use);
				}

			}

		  sentence_delete(sent);
		  sprintf(message,"</linkgram>\r\n");
		  sock_print(sock,message);
		  writeOutputTrailer(sock);
		  //comm_close_socket(active_socket);

		  //------------------------------------------------
          //char * cbuf = do_work(recvBuffer, totalReceived);


        }
        totalReceived = 0;  // MLW 9/7/98
		// Note to log file
		flog = fopen("flog.txt","a");
		if (flog!=NULL)
		{
		fprintf(flog,"<transact>\r\n");
		fprintf(flog,"<origsent>%s</origsent>\r\n",recvBuffer);
		fprintf(flog,"<wordcount>%i</wordcount>\r\n<parsecount>%i</parsecount>\r\n<totalparses>%i</totalparses>\r\n",count_to,num_to_query,total_valid_linkages);
	    fprintf(flog,"<sourceip> %s</sourceip> ",(char *)inet_ntoa(clientAddr.sin_addr));
	    if (entity!=NULL) fprintf(flog,"<sourcedomain>%s</sourcedomain>",entity->h_name );
	    fprintf(flog,"\n");

		fprintf(flog,"</transact>\r\n");
		
		fclose(flog);
		}

		// Close the lexical XML cache
		if (extrakbfile!=NULL) fclose(extrakbfile);
		extrakbfile=NULL;


      } while (totalReceived > 0);

      //if (DEBUG) cout << "Connection closed by client." << endl;
      close(sock);
	  printf("render_service:comm_close_socket\n");
    }
    WSACleanup();

}
else
{
    while (fget_input_string(input_string, stdin, stdout, opts)) {

	if (space_in_use != dictionary_and_option_space + reported_leak) {
	    fprintf(stderr, "Warning: %d bytes of space leaked.\n",
		    space_in_use-dictionary_and_option_space-reported_leak);
	    reported_leak = space_in_use - dictionary_and_option_space;
	}

	if ((strcmp(input_string, "quit\n")==0) ||
	    (strcmp(input_string, "exit\n")==0)) break;

	if (special_command(input_string, dict)) continue;
	if (parse_options_get_echo_on(opts)) {
	    printf("%s", input_string);
	}

	if (parse_options_get_batch_mode(opts)) {
	    label = strip_off_label(input_string);
	}

	sent = sentence_create(input_string, dict);

	if (sent == NULL) {
	    if (verbosity > 0) fprintf(stderr, "%s\n", lperrmsg);
	    if (lperrno != NOTINDICT) exit(-1);
	    else continue;
	} 
	if (sentence_length(sent) > parse_options_get_max_sentence_length(opts)) {
	    sentence_delete(sent);
	    if (verbosity > 0) {
	      fprintf(stdout, 
		      "Sentence length (%d words) exceeds maximum allowable (%d words)\n",
		    sentence_length(sent), parse_options_get_max_sentence_length(opts));
	    }
	    continue;
	}

	/* First parse with cost 0 or 1 and no null links */
	parse_options_set_disjunct_cost(opts, 2);
	parse_options_set_min_null_count(opts, 0);
	parse_options_set_max_null_count(opts, 0);
	parse_options_reset_resources(opts);

	num_linkages = sentence_parse(sent, opts);

	/* Now parse with null links */
	if ((num_linkages == 0) && (!parse_options_get_batch_mode(opts))) {
	    if (verbosity > 0) fprintf(stdout, "No complete linkages found.\n");
	    if (parse_options_get_allow_null(opts)) {
		parse_options_set_min_null_count(opts, 1);
		parse_options_set_max_null_count(opts, sentence_length(sent));
		num_linkages = sentence_parse(sent, opts);
	    }
	}

	if (parse_options_timer_expired(opts)) {
	    if (verbosity > 0) fprintf(stdout, "Timer is expired!\n");
	}
	if (parse_options_memory_exhausted(opts)) {
	    if (verbosity > 0) fprintf(stdout, "Memory is exhausted!\n");
	}

	if ((num_linkages == 0) && 
	    parse_options_resources_exhausted(opts) &&
	    parse_options_get_panic_mode(opts)) {
	    print_total_time(opts);
	    if (verbosity > 0) fprintf(stdout, "Entering \"panic\" mode...\n");
	    parse_options_reset_resources(panic_parse_opts);
	    parse_options_set_verbosity(panic_parse_opts, verbosity);
	    num_linkages = sentence_parse(sent, panic_parse_opts);
	    if (parse_options_timer_expired(panic_parse_opts)) {
		if (verbosity > 0) fprintf(stdout, "Timer is expired!\n");
	    }
	}

	print_total_time(opts);

	if (parse_options_get_batch_mode(opts)) {
	    batch_process_some_linkages(label, sent, opts);
	}
	else {
	    process_some_linkages(sent, opts);
	}

	sentence_delete(sent);
	if (external_space_in_use != 0) {
	    fprintf(stderr, "Warning: %d bytes of external space leaked.\n", 
		    external_space_in_use);
	}
    }
}
    if (parse_options_get_batch_mode(opts)) {
	print_time(opts, "Total");
	fprintf(stderr, 
		"%d error%s.\n", batch_errors, (batch_errors==1) ? "" : "s");
    }

    parsing_space_leaked = space_in_use - dictionary_and_option_space;
    if (parsing_space_leaked != 0) {
        fprintf(stderr, "Warning: %d bytes of space leaked during parsing.\n", 
		parsing_space_leaked);
    }

    parse_options_delete(panic_parse_opts);
    parse_options_delete(opts);
    dictionary_delete(dict);

    if (space_in_use != parsing_space_leaked) {
        fprintf(stderr, 
		"Warning: %d bytes of dictionary and option space leaked.\n", 
		space_in_use - parsing_space_leaked);
    } 
    else if (parsing_space_leaked == 0) {
        fprintf(stderr, "Good news: no space leaked.\n");
    }

    if (external_space_in_use != 0) {
        fprintf(stderr, "Warning: %d bytes of external space leaked.\n", 
		external_space_in_use);
    }

    return 0;
}
